/* raw_io.c  -- Write out a heightfield as triangle vertices in
 *              RAW ascii format. This is horribly inefficient as
 *              far as file space goes, but does allow an interface
 *              to polygon-only based rendering tools
 *              John P. Beale 4/27/96
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdarg.h>
#include "hl.h"

#define U unsigned int

int raw_write(hfield hf, char *fname)
{
  int xsize, ysize;
  int xs, ys;
  int ix, iy;
  FILE *outfile;
  double x1,y1,z1,x2,y2,z2,z3,z4;  /* vertices of two triangles */

    /* ------------- open RAW (ascii vertexes) file ----------- */

  xsize = hf.xsize;
  ysize = hf.ysize;
  xs = xsize - 1;
  ys = ysize - 1;
  outfile = fopen(fname, "wb");
  if (outfile == NULL) {
    fprintf(stderr,"Trouble opening RAW file %s\n",fname);
    return(1);
  }

    /*----------- Write out data to RAW format file ------- */
    fprintf(outfile,"%s\n",hf.name); /* first line: HF data name */
    for (iy = 0; iy < ys; iy++) {
     for (ix = 0; ix < xs; ix++) {
       x1 = (D)ix/xs; y1 = (D)iy/ys;
       x2 = (D)(ix+1)/xs; y2 = (D)(iy+1)/ys;
       z1 = (D)El(hf.a,ix,iy);
       z2 = (D)El(hf.a,ix+1,iy);
       z3 = (D)El(hf.a,ix,iy+1);
       z4 = (D)El(hf.a,ix+1,iy+1);
       fprintf(outfile,"%1.4f %1.4f %1.4f",x1,y1,z1);
       fprintf(outfile," %1.4f %1.4f %1.4f",x2,y1,z2);
       fprintf(outfile," %1.4f %1.4f %1.4f\n",x2,y2,z4); /* tri #1 */
       fprintf(outfile,"%1.4f %1.4f %1.4f",x1,y1,z1);
       fprintf(outfile," %1.4f %1.4f %1.4f",x2,y2,z4);
       fprintf(outfile," %1.4f %1.4f %1.4f\n",x1,y2,z3); /* tri #2 */
     }
    }
  
  fclose(outfile);
  return(0);

} /* end raw_write() */

