/* draw a heightfield on the screen using libgrx (DOS djgpp graphics) */
/* JPB 9/27/95 */

#include <grx20.h>                      /* djgpp graphics library */
#include <stdio.h>                      /* fopen() sprintf() etc. */
#include <stdlib.h>                     /* strtod() */
#include <math.h>                       /* math stuff */
#include <string.h>                     /* strcpy() */
#include <ctype.h>                      /* tolower() */
#include <pc.h>                         /* non-portable: kbhit() */
#include <conio.h>                      /* non-portable: clrscr(), kbhit() */
#include <mousex.h>                     /* mouse cursor */
#include "hl.h" 
#include "hf_io.h"

#define ESC '/033'

#define TCOLOR 7                /* color of text on screen */
#define FALSE 0
#define TRUE 1
#define D double
#undef tolower             /* annoying ANSI detail */                           
			   /* wrap around coords */

		    /* why aren't these in the conio.h file!? */
void clrscr(void);   
int kbhit(void);

		/* draw two triangles joined along hypotenuse */
void tri3d(D x0, D z0,D x1,D z1,D y0,D y1,D y2,D y3, int s);
void drawhf(PTYPE *d1, int maxx, int maxz, int s, int t);
int graph_setup(int graph_mode, int *xsize, int *ysize);
void graph_close();   /* close graphics screen */
void sgraph_close();  /* close graph only if not already in view mode */
void graph_erase();   /* erase graphics screen */
void pr_shelp();              /* print help message */
void pr_vars();              /* print variable values */

int show_grid(hfield hf, char inip[MAXTOK][TOKLEN], int itoks);
int show_flat(hfield hf, char inip[MAXTOK][TOKLEN], int itoks);
int h_plots(hfield hf, int x1,int y1,int x2, int y2);
int ss_flat(int show4,int xgsize, int ygsize);


extern int getline(char s[], int lim);       /* read in one line */

extern hfield *hst;      /* stack of heightfields */
extern int hp;           /* pointer to top of stack: hst[hp] */
extern int stacksize;    /* size of stack */
extern int nskip;        /* number of hf elements to skip */
extern int v_tile;       /* view as a tiled surface? */
extern int graph_mode;   /* initial graphics resolution mode */
extern int logging;      /* if command logging is turned on */
extern int takefile;     /* if a command file is being read in */
extern int echo_cmd;     /* if commands from take file should be echoed */
extern int hist_on;      /* record commands? */
extern int view_current;   /* if currently in view mode */
extern int show_current;   /* if currently in show mode */
extern FILE *logfp;            /* command logging file */
extern FILE *cmd_stream;       /* input stream (stdin or command file) */
extern D sl_min;         /* minimum slope (shading offset) */
extern D sl_scale;       /* slope scale factor (shading scale) */
extern int_done;         /* if we're ready to quit program */
extern PT3 eye;          /* viewpoint */
extern PT3 light;        /* illumination vector */

static int xgsize,ygsize;  /* dimensions of graphics screen for 'show' */

D zcenter;
D zfar, znear;
D zsf;      /* z coordinate scale factor */
D hs;      /* vertical (y) coordinate scale factor */
D ycorr;  /* y offset correction */
D zeps;   /* z epsilon */
int yoff;  /* screen coords y-offset */
int sxmid;  /* center coordinates of view window (eg 640x480) */
int symid;
int maxcol = 255;  /* maximum color index */
struct Point3 light;  /* position of light source */
D bright;           /* intensity of light  */
int maxcon;         /* maximize contrast? (boolean) */
int cmin, cmax;     /* max and min color index value in this image */
int erasewin = TRUE;       /* erase window between displays? */
int done1 = FALSE;
int repaint = TRUE;    /* whether 'view' redraws 3D projection or not */

 /* create 3D projection of heightfield on graphics screen */
int show_grid(hfield junk, char inip[MAXTOK][TOKLEN], int itoks)
{
  hfield hf;
  PTYPE *d1;
  int xsize, ysize;
  char insp[MAXTOK][TOKLEN];  /* for internal command-line parser */
  int tokens;
  int len;                    /* overall length of input line */
  int xgsize=0, ygsize=0; /* screen size */
  int tt;         /* index to tokens */
  char c,d;         /* first,second char of cmd */
  int i;
  int temp_int;
  int grid_cmd;   /* if line is 'grid' command */
  int first_l;    /* if this is first prompt. oh what a kluge. */
  D eyinc;
  D lmag;         /* magnitude of a vector */
  int shader;     /* 0 wireframe 1 height 2 slope 3 lightsource */
  char instr[CLINELEN];  /* interpret raw string of commands */
  int t_dur= 0;        /* duration of pause time */

  if (s_empty()) return(1);
  hf = hst[hp];
  xsize = hst[hp].xsize;
  ysize = hst[hp].ysize;
  hs = 0.3;      /* vertical (y) coordinate scale factor */

  tt = 1;
  if (itoks > 1) {
   do {
    if (tolower(inip[tt][0])=='p') sscanf(&inip[tt][1],"%d",&nskip);
    else if (tolower(inip[tt][0])=='w') sscanf(&inip[tt][1],"%d",&t_dur);
    else if (tolower(inip[tt][0])=='s') sscanf(&inip[tt][1],"%d",&shader);
    else if (tolower(inip[tt][0])=='t') sscanf(&inip[tt][1],"%d",&v_tile);
    else if (tolower(inip[tt][0])=='g') 
      sscanf(&inip[tt][1],"%d",&graph_mode);
    tt++;
    } while (tt < itoks);
   } /* end if itoks>1 */

  if (t_dur<0) t_dur=0;        /* sanity checks; force correct range */
  if (nskip<1) nskip=1;
  if (shader > 3) shader = 3;
  if (shader < 0) shader = 0;
  if (graph_mode>3) graph_mode = 3;

  /* nskip = (xsize / 50) + 1; */
  /* getch(); */
  MouseEventEnable(1,0);        /* enable keyboard events */
  graph_setup(graph_mode, &xgsize,&ygsize);
  view_current=TRUE;

  sxmid = xgsize / 2;  /* middle of screen in screen coords */
  symid = ygsize / 2;
  zsf = 0.3;      /* z coordinate scale factor */
  ycorr = 0.0;  /* y offset correction */
  zeps = 0.001;

  eyinc = 0.8;
  lmag = sqrt(light.x*light.x + light.y*light.y + light.z*light.z);
  light.x /= lmag;
  light.y /= lmag;
  light.z /= lmag;

  bright = 0.8;  /* lightsource intensity */
  zcenter = 1;
  shader = 3;    /* shading model: wireframe, angle, light, etc. */

  repaint = TRUE;
  maxcon = FALSE;
  cmin = 0; cmax = maxcol;
  done1 = FALSE;                /* haven't drawn a figure yet */
  /* ------------------------------------main interp. loop ------- */
  do {
    zcenter = -eye.z;
    ycorr = 2*eye.y/(zcenter);
    yoff = (int) (symid*(ycorr + eye.y/zcenter));
    if (repaint) {
      if (erasewin && done1) graph_erase();
      done1 = TRUE;                             /* draw one screen yet? */
      if( ( d1 = hst[hp].a) == (PTYPE *)NULL) {
	  graph_close();          /* close graphics window */  
	  printf("No matrix to view.\n"); 
	  return(1);
      }
      xsize = hst[hp].xsize;
      ysize = hst[hp].ysize;
      
      graph_erase();  /* clear old stuff off screen */
      clrscr();                 /* clear text screen */
      printf("          ?: help  Q: quit\n");
      drawhf(hst[hp].a,hst[hp].xsize,hst[hp].ysize,shader,v_tile);
      if (t_dur > 0) {
	sleep((U)t_dur);        /* pause for spec'd # of seconds */
	graph_close();          /* close graphics window */  
	return(0);
      }
    }

    repaint = TRUE;
    
    /* ------------------------------------main interp. loop ------- */
    /* hist_on = FALSE; */
    first_l = TRUE;              /* first prompt since clearing */
    do {
      printf("view>");            /* input prompt for user entry */
      getline(instr,80);
      if (instr[0]=='\0')  {
	 graph_erase();  /* clear old stuff off screen */
	 clrscr();                 /* clear text screen */
	 printf("          ?: help  Q: quit\n");
	 drawhf(hst[hp].a,hst[hp].xsize,hst[hp].ysize,shader,v_tile);
       }
    } while (instr[0] == '\0'); /* now have non-null line */
    /* hist_on = TRUE; */
    
    c = tolower(instr[0]);
    d = tolower(instr[1]);
    len = strlen(instr);  
    grid_cmd = FALSE;
    if (len < 3) {
     grid_cmd = TRUE;
     if (c == '+') hs *= 1.3;
     else if (c == '-') hs /= 1.3;
     else if (c == 'u') eye.y -= eyinc;
     else if (c == 'd') eye.y += eyinc;
     else if (c == 'b') eye.z -= 1.0;
     else if (c == 'f') {
       eye.z += 0.5;
       if (eye.z >= -1.1) eye.z = -1.1;
     }
     else if (c == '>') bright *= 1.2;
     else if (c == '<') bright /= 1.2;
     else if (c == 'm') maxcon = !maxcon;
     else if (c == 'v') erasewin = TRUE;
     else if ((c == 'p') && (len==1) ) nskip=1;
     else if ((c == 'p') && (tolower(instr[1])=='v') )
      {
	pr_vars();
	repaint = FALSE;
      }
     else if (c == 'g' && d>='0' && d<='9') {
       sscanf(&instr[1],"%d",&graph_mode);
       graph_setup(graph_mode, &xgsize,&ygsize);
       sxmid = xgsize / 2;  /* middle of screen in screen coords */
       symid = ygsize / 2;
     }
     else if ((c == 'h') || (c == '?')) {
       pr_shelp();
       repaint = FALSE;
     }
     else if (c == 't') v_tile = !(v_tile);
     else if ((c == 's')&&(d>='0'&&d<='9')) {
      maxcon = FALSE;
      shader = (int)atol(&instr[1]);
      shader--;
      if (shader > 3) shader = 3;
      if (shader < 0) shader = 0;
     }
     else if (c == 'q') { /* quitting */ }
     else grid_cmd = FALSE;
    } /* endif (len<3) */

    if (!grid_cmd) {
     if (instr[0] == 'L') {
       maxcon = FALSE;      /* reset contrast limits */
       sscanf(&instr[1],"%lf %lf %lf",&light.x,&light.y,&light.z);
       lmag = sqrt(light.x*light.x + light.y*light.y + light.z*light.z);
       light.x /= lmag;
       light.y /= lmag;
       light.z /= lmag;
     }
     else if (instr[0] == 'E') {
       maxcon = FALSE;      /* reset contrast limits */
       sscanf(&instr[1],"%lf %lf %lf",&eye.x,&eye.y,&eye.z);
       if (eye.z > -1.01) eye.z = -1.01;
     }
     else {
       repaint = FALSE;
       
       parse_line(instr,&tokens,insp);
       /* printf("read %d tokens %s\n",tokens,insp[0]); */
       if ((c == 'p') && (tokens == 1) && 
	  ((temp_int = (int)atol(&instr[1])) > 0) )   {
				       /* oops, grid cmd after all */
	 nskip = temp_int;
	 if (nskip < 1) nskip = 1;
	 repaint = TRUE;
       } else cmd_seq1(instr);  /* interpret raw string of commands */
       
       int_done = FALSE;      /* don't let 'quit' trigger total exit */
       for (i=0;i<tokens;i++) insp[i][0] = '\0';  /* erase the old tokens */
     }
    }  /* not grid_cmd */
  } while (c != 'q');

  graph_close();          /* close graphics window */  
  return(0);
}

void pr_vars()              /* print variable values */
{
   printf("eye coords:   %f %f %f\n",eye.x,eye.y,eye.z);
   printf("light vector: %f %f %f\n",light.x,light.y,light.z);
   printf("vertical scale: %f\n",hs);
   printf("lightsource I:  %f\n",bright);
   printf("contrast maximization is %s.\n",maxcon ? "ON" : "OFF");
   printf("color min,max: %d - %d\n",cmax,cmin);
   printf("grid spacing P%d",nskip);
   printf(" clear is %s.\n",erasewin ? "ON":"OFF");
   printf("\n");
}


void pr_shelp()              /* print help message */
{
    printf("----------- View Commands ------------------------\n");
    printf("---  There are four shading options as follows:\n");
    printf("S1  wireframe   S2  height-shading\n");
    printf("S3  slope       S4  lightsource\n");
    printf("L x y z   vector to lightsource (must be capital 'L')\n");
    printf("E x y z   position of viewer's eye (must be capital 'E')\n");
    printf("B/F   move eye position back/forward.\n");
    printf("U/D   move eye position up/down\n");
    printf("M     toggle 'maximize contrast' mode\n");
    printf(">/<   brighter/darker (S4 mode only)\n");
    printf("+/-   change vertical scaling of heightfield.\n");
    printf("T     tile heightfield in a 2x3 array\n");
    printf("Pn    draw only every nth line, for faster preview.\n");
    printf("Gn    graphic mode, G1=640x480 G2=800x600 G3=1024x768\n");
    printf("V     view / redraw the display\n");
    printf("PV    print variables   ? help\n");
    printf("Q     quit\n");
    printf("\n");
}

/* =============================================================== 
 * project two triangles, given 3D coordinates, on a 2D viewscreen 
 * uses gobal vars:
 *   hs : vertical heightfield scaling factor
 *   eye.x,y,z : location of eye (viewpoint)
 *   symid : y coordinate midpoint
 *   ycorr : y correction factor
 *   yoff  : y offset in screen units 
 *   zcenter : center to 'rotate' appparent hf around
 *   maxcol  : maximum color value 
 *   bright  : lightsource intensity 
 * ===============================================================
 */

/*  shader: 0 = wireframe, 1 = height color, 2 = slope, 3 = lightsource */

void tri3d(D x0, D z0,D x1,D z1,D y0,D y1,D y2,D y3, int shader)
{
 D zmod0, zmod1; 
 D slope1;
 D slope2; 
 int c1, c2;  /* colors */
 D nx,ny,nz,nmag;   /* normal vector */
 D ctheta;          /* cos(theta)  theta = angle between light and normal */

 int p[3][2];       /* three sets of points in 2D */
 c2 = c1 = 0;            /* make compiler happy */ 

 if (shader == 1) {  /* height, before transform */
   c1 = maxcol*(y0 + y1 + y2 + y3)/4;
   c2 = c1;
 } 
 
 /* x0,y0,z0 should be in the range -1..1 */
 /* transform to eye coordinate (eye.x,y,z) */
 
 y0 *= -hs; y1 *= -hs; y2 *= -hs; y3 *= -hs;                          
 x0 -= eye.x; y0 -= eye.y; z0 -= eye.z;
 x1 -= eye.x; y1 -= eye.y; z1 -= eye.z;
 y2 -= eye.y; y3 -= eye.y; 
 if ((z1 <= zeps) || (z0 <= zeps)) return;  /* foreground clipping */

 if (shader == 2) {  /* shading based on slope */
   slope1 = ATAN2( (z0-z1)*(x1-x0),(z1-z0)*(y2-y0) );
   c1 = (int) maxcol*(slope1-sl_min)*sl_scale;
  
   slope2 = ATAN2( (z0-z1)*(x1-x0),(z1-z0)*(y1-y3) ); 
   c2 = (int) maxcol*(slope2-sl_min)*sl_scale;
   
 } else if (shader == 3) { /* compute normal, dot product with lightsource */

   nx = (z0-z1)*(y2-y0);
   ny = (z1-z0)*(x1-x0);
   nz = (x1-x0)*(y2-y1);
   nmag = sqrt(nx*nx + ny*ny + nz*nz);
   ctheta = (nx*light.x + ny*light.y + nz*light.z) / nmag;
   c1 = (int) (maxcol>>1)*(bright*(ctheta + 1.0));
   
   nx = -(z0-z1)*(y3-y1);
   /* ny = (z1-z0)*(x1-x0); */   /* duplicates line above */
   nz = (x0-x1)*(y3-y0);
   nmag = sqrt(nx*nx + ny*ny + nz*nz);
   ctheta = (nx*light.x + ny*light.y + nz*light.z) / nmag;
   c2 = (int) (maxcol>>1)*(bright*(ctheta + 1.0));

 } else if (shader == 0) {  /* shader = 0: wireframe */
   c1 = maxcol;  /* white, we hope. brightest color, anyway */
 }

 if (maxcon &&(cmax!=cmin)) {          /* if maximize_contrast is TRUE... */
   c1 = (int) (((D)maxcol/(cmax-cmin))*(c1-cmin));
   c2 = (int) (((D)maxcol/(cmax-cmin))*(c2-cmin));
 } 
 if (c1 < 0) c1 = 0;
 if (c1 > maxcol) c1 = maxcol;
 if (c2 < 0) c2 = 0;
 if (c2 > maxcol) c2 = maxcol;

 if (!maxcon) {
   if (c1 > cmax) cmax = c1;
   if (c1 < cmin) cmin = c1;
 }
 if (c1 == TCOLOR) c1 = TCOLOR-1;  /* avoid text color: KLUGE! */
 if (c2 == TCOLOR) c2 = TCOLOR-1;

 zmod0 = z0 * zsf;   /* z1 multiplied by z scale factor */
 zmod1 = z1 * zsf;
/*
 printf("zmod: %f %f\n",zmod0,zmod1);
 printf("color %d %d\n",c1,c2);
 getch(); */
  /* sxmid, symid are midpoints of screen, eg sxmid=320 for 640x480 scn. */

 p[0][0] = (int) (sxmid*x0/zmod0) + sxmid;  /* x,y,z 0 */
 p[0][1] = (int) (symid*y0/zmod0) + symid + yoff;
 
 p[1][0] = (int) (sxmid*x1/zmod1) + sxmid;  /* x,y,z 1 */
 p[1][1] = (int) (symid*y1/zmod1) + symid + yoff;

 p[2][0] = (int) (sxmid*x1/zmod0) + sxmid; /* x,y,z 2 */
 p[2][1] = (int) (symid*y2/zmod0) + symid + yoff;
/* printf("%d %d %d %d %d %d : %d %d %d\n",
   p[0][0],p[0][1],p[1][0],p[1][1],p[2][0],p[2][1],c1,c2,maxcol/2); 
  */
 if( ((p[0][0]==p[1][0]) && (p[0][1]==p[1][1])) ||
     ((p[0][0]==p[2][0]) && (p[0][1]==p[2][1])) ||
     ((p[1][0]==p[2][0]) && (p[1][1]==p[2][1]))  )  {
       GrPlot(p[0][0],p[0][1],c1);
   } else {
     if (shader == 0) {
       GrFilledConvexPolygon(3,p,0);
       GrPolygon(3,p,maxcol);
     }
     else if (shader == 1) {
       GrFilledConvexPolygon(3,p,c1);
       GrPolygon(3,p,maxcol/2+c1/2);
     } else /* shader =2,3: slope or lightsource shading */
       GrFilledConvexPolygon(3,p,c1);   
   }

 p[1][0] = (int) (sxmid*x0/zmod0) + sxmid;  /* x,y,z 0 */
 p[1][1] = (int) (symid*y0/zmod0) + symid + yoff;

 p[0][0] = (int) (sxmid*x1/zmod1) + sxmid;  /* x,y,z 1 */
 p[0][1] = (int) (symid*y1/zmod1) + symid + yoff;

 p[2][0] = (int) (sxmid*x0/zmod1) + sxmid; /* x,y,z 3 */
 p[2][1] = (int) (symid*y3/zmod1) + symid + yoff; 

if( ((p[0][0]==p[1][0]) && (p[0][1]==p[1][1])) ||
     ((p[0][0]==p[2][0]) && (p[0][1]==p[2][1])) ||
     ((p[1][0]==p[2][0]) && (p[1][1]==p[2][1]))  )  {
       GrPlot(p[0][0],p[0][1],c2);
   } else {
     if (shader == 0) {
       GrFilledConvexPolygon(3,p,0);
       GrPolygon(3,p,maxcol);
     }
     else if (shader == 1) {
       GrFilledConvexPolygon(3,p,c2);
       GrPolygon(3,p,maxcol/2+c2/2);
     } else 
       GrFilledConvexPolygon(3,p,c2);   
   }
       
} /* end tri3d() */

/* draw a heightfield, back to front, as a series of triangles */
/* uses global variable nskip as hf index increment. */

void drawhf(PTYPE *d1, int xsize, int ysize, int shader, int tile)
{
  int ix,iz,izi;
  int ymin, ymax, xmin, xmax;
  int noffx, noffz;
  D xd2, yd2;
  D x0, y0, z0, x1, y1, z1;
  D y2, y3;
  char c;  

  xd2 = xsize / 2.0;
  yd2 = ysize / 2.0;
  if (tile) {
    ymin = 0;
    ymax = (3*ysize) - 1;
    xmin = -xsize + 1;
    xmax = 2*xsize -1;
  } else {
    ymin = 0;
    ymax = ysize-1;
    xmin = 0;
    xmax = xsize-1;
  }
  if (!maxcon) {
    cmax = 0; 
    cmin = maxcol;     /* initialize global lims to extrema */
  }

		     /* loop over the two array indices */        

    for ( iz=ymin;iz<ymax;iz += nskip) {
	  izi = (ymax - iz) + ymin;
	  if ((iz+nskip) < ymax) noffz = nskip;
	     else noffz = ymax-iz;
	  for ( ix=xmin;ix<xmax;ix+= nskip) {
	    if ((ix+nskip) < xmax) noffx = nskip;
	       else noffx = xmax-ix;
	    x0 = (ix-xd2)/xd2; z0 = (izi-yd2)/yd2; 
	    x1 = (ix+noffx-xd2)/xd2; z1 = (izi-noffz-yd2)/yd2; 
	    y0 = (D) Elmod(d1,ix,iz);
	    y1 = (D) Elmod(d1,ix+noffx,iz+noffz);
	    y2 = (D) Elmod(d1,ix+noffx,iz);
	    y3 = (D) Elmod(d1,ix,iz+noffz);
	    tri3d(x0,z0,x1,z1,y0,y1,y2,y3,shader); 
	    if (kbhit()) {  /* escape from loop if key hit */  
	      c=getch();
	      return;  
	    }              
	  } /* end for iz */
    } /* end for ix */

} /* end drawhf() */


int graph_setup(int mode, int *xsize, int *ysize)        
{
int i;
int red,green,blue;

/*        atype = GrAdapterType();
	  printf("Graphics adaper type = %d\n",atype);
	  mtype = getch(); */
	
	if (mode == 0)
	  GrSetMode(GR_width_height_color_graphics,320,200,256);
	else if (mode == 1)
	  GrSetMode(GR_width_height_color_graphics,640,480,256);
	else if (mode == 2)
	  GrSetMode(GR_width_height_color_graphics,800,600,256);
	else   /* mode == 3 */
	  GrSetMode(GR_width_height_color_graphics,1024,768,256);

	xsize[0] = GrSizeX();
	ysize[0] = GrSizeY();

	GrResetColors();
	for (i=0;i<256;i++) {
	  red = i;
	  green = i+3;
	  if (green > 255) green = 255;
	  blue = i;
	  if (i==TCOLOR) red=green=blue=180;  /* color of text printf() */
	  GrSetColor(i,red,green,blue);
	}
    return (0);
} /* end graph_setup */

void graph_close()
{
  MouseEventEnable(0,0); /* enable keyboard, mouse events */
  GrSetMode(GR_default_text); 
  view_current = FALSE;
}

void sgraph_close()  /* close graph only if not already in view mode */
{
  if (!view_current) {
    MouseEventEnable(0,0); /* enable keyboard, mouse events */
    GrSetMode(GR_default_text); 
  } else {
    graph_erase();  /* must call twice due to bug in VESA driver (?) */
    graph_erase();
    repaint = TRUE;        /* redraw the 3d view after returning */
  }
}

void graph_erase()
{
  GrClearScreen(0);      /* clear graphics screen */
  clrscr();              /* clear text screen */
  GrClearScreen(0);      /* clear screen again. arghh. */
  if (!(takefile && !echo_cmd))
   { if (graph_mode==0) printf("\n\n\n\n\n\n");
     else printf("\n\n\n\n");       /* move down four lines */
   }
}

		/* show flat 2D image as greyscale */
int show_flat(hfield junk, char inip[MAXTOK][TOKLEN], int itoks)
{
 char instr[CLINELEN];               /* input command line */
 char insp[MAXTOK][TOKLEN];  /* for internal command-line parser */
 int i,tokens;
 int xsize, ysize;
 hfield hf;
 int show4 = FALSE;
 int gr_clear = TRUE;
 int t_dur = 0;
 int tt;
 int show_exit = FALSE;         /* exit from 'show' mode */
 /* MouseEvent m_event;            */  /* mouse handling */

 hf = hst[hp];
 if( hf.a == (PTYPE *)NULL) {
   printf("No matrix to show.\n");
   return(1);
 }

 /* 
    for (tt=0;tt<itoks;tt++) printf("%s",inip[tt]);
    printf("\n%d tokens total\n",itoks);
    getch();
 */

  tt = 1;
  if (itoks > 1) {
    do {
      if (tolower(inip[tt][0])=='p') sscanf(&inip[tt][1],"%d",&nskip);
      else if (keymatch(inip[tt],"nc",2)) gr_clear = FALSE;
      else if (tolower(inip[tt][0])=='w') sscanf(&inip[tt][1],"%d",&t_dur);
      else if (tolower(inip[tt][0])=='t') sscanf(&inip[tt][1],"%d",&v_tile);
      else if (tolower(inip[tt][0])=='4') show4 = TRUE;
      else if (tolower(inip[tt][0])=='g') 
	sscanf(&inip[tt][1],"%d",&graph_mode);
      else printf("show: %s not understood.\n",inip[tt]);
      tt++;
    } while (tt < itoks);
  }
  
  if (graph_mode>3) graph_mode = 3;
  if (graph_mode<0) graph_mode = 0;

  if (t_dur<0) t_dur=0;        /* sanity checks; force correct range */
  if (nskip<1) nskip=1;

 xsize = hf.xsize;
 ysize = hf.ysize;
/* if (hf.max==hf.min && !hf.c) {
   printf("Constant matrix (min = max = %5.3f).\n",hf.min);
   return(1);
 }
 */
 if (!show_current) graph_setup(graph_mode, &xgsize,&ygsize);
 repaint = TRUE;     /* be sure to draw it the first time */
 do {
    if (repaint) { 
      clrscr();                 /* clear text screen */
      if (gr_clear) {
	graph_erase();  /* clear old stuff off screen */
      }
      ss_flat(show4,xgsize,ygsize); /* show 1 or 4 HFs */
      repaint = FALSE;
      gr_clear = TRUE;
    }

    if (show_current || (t_dur > 0)) {
	if (t_dur > 0) sleep((U)t_dur);      /* pause for this # seconds */
	if (!show_current) sgraph_close();   /* close graphics window */  
	return(0);
    }
    
    do {                           /* get a non-null cmd line */
       if (!(takefile && !echo_cmd))
	printf("show>");           /* input prompt for user entry */
       getline(instr,CLINELEN-1);  /* get one line of command(s) */
       if (instr[0]=='\0') {       /* return repaints */
	 graph_erase();  /* clear old stuff off screen */
	 ss_flat(show4,xgsize,ygsize); /* show 1 or 4 HFs */
       }
    } while (instr[0] == '\0');    /* now have non-null line */

    parse_line(instr,&tokens,insp);
    if (keymatch(insp[0],"show",1)) {
       repaint=TRUE;
       if (keymatch(insp[1],"4",1)) show4 = TRUE;
	 else show4 = FALSE;
       if (keymatch(insp[1],"nc",2) || keymatch(insp[2],"nc",2)) 
	 gr_clear = FALSE;
    } else if (keymatch(insp[0],"view",1)) {
      printf("Quit 'show' before entering VIEW.\n");
    } else if (keymatch(insp[0],"quit",1)) { /* quit */
      show_exit = TRUE;
    } else {
      show_current = TRUE;  /* we're now in interactive show mode */
      cmd_seq1(instr);  /* interpret entire string of command(s) */
      show_current = FALSE; /* don't exit at top of loop now */
    }
    for (i=0;i<tokens;i++) insp[i][0] = '\0';  /* erase the old tokens */
 } while (show_exit != TRUE);
 
 /*
 if (instr[0]=='m') {
   MouseEventMode(1);
   MouseInit();
   MouseEventEnable(1,1); 
   do { MouseDisplayCursor();
	MouseGetEvent(M_KEYPRESS,&m_event);
    } while (m_event.kbstat == 0);
   MouseEraseCursor();
   MouseUnInit();
 } 
 */

 sgraph_close();          /* close graphics window */  
 show_current = FALSE;    /* we're not in interactive 'show' mode anymore */
 return(0);
} /* s_flat() */

 /* show hf on screen from screen coords x1,y1 to x2,y2 */
int h_plots(hfield hf, int x1,int y1,int x2, int y2)
{ 
int xsize,ysize;
D xsf,ysf,cscale;
D hmin,hmax;
D imin,imax;
int x,y,xx,yy,c;
int xoff, yoff;        /* offset in screen coordinates */
int xrange,yrange;     /* range of screen coordinates */
 
 if (hf.a==NULL) return(1);
 xsize = hf.xsize;
 ysize = hf.ysize;
 xrange = x2-x1;
 yrange = y2-y1;
 
 for (x=0;x<1000;x++)
   c=GrPixel(x2,y2);    /* get pixel color at x,y  */
/*
   printf("%dx%d at %d,%d..%d,%d min=%f max=%f c=%d\n", 
   xsize,ysize,x1,y1,x2,y2,hf.min,hf.max,c); 
   getch(); 
*/
 if (hf.c) {    /* display complex matrix as two parts */
  yoff = ( (y2-y1) - ((x2-x1)/2) ) / 2;
  y1 += yoff;
  y2 -= yoff;
  xsf = (D)xsize/(x2-x1);
  ysf = (D)ysize/(y2-y1);
  r_minmax(hf,&hmin,&hmax);
  if (hmax != hmin)
   cscale = 255.0 / (hmax - hmin);
  else cscale = 0;

  for (y=y1;y<y2;y++) {    /* display the real part */
   yy = (int)((y-y1)*ysf);
   for (x=x1;x<x2;x++) {
    xx = (int)((x-x1)*xsf);
    if (cscale==0) c=128;
    else c = (int) ((El(hf.a,xx,yy) - hmin) * cscale);
    if (c == TCOLOR) c--;
    GrPlot((x+x1)/2,y,c);
   } /* for y */
  } /* for x */
  
  i_minmax(hf,&imin,&imax);  /* display complex part */
  if (imax != imin) cscale = 255.0 / (imax - imin);
   else cscale = 0;
  xoff = (x2-x1)/2.0;
  for (y=y1;y<y2;y++) {
   yy = (int)((y-y1)*ysf);
   for (x=x1;x<x2;x++) {
    xx = (int)((x-x1)*xsf);
    if (cscale == 0) c = 128;
     else c = (int) ((Im(hf.a,xx,yy) - imin) * cscale);
    if (c == TCOLOR) c--;
    GrPlot((x+x1)/2+xoff,y,c);
   } /* for y */
  } /* for x */

 } else {                       /* HF is not complex */
  if (xrange != yrange) {
    xoff = (MAX(xrange,yrange) - MIN(xrange,yrange));
    if (xrange>yrange) {   /* (always true on PC screen...) */
      x1 += xoff/2;
      x2 -= xoff/2;
    } else {
      y1 += xoff/2;
      y2 -= xoff/2;
    }
  }
  xsf = (D)xsize/(x2-x1);
  ysf = (D)ysize/(y2-y1);

  r_minmax(hf,&hmin,&hmax);
  if (hmax != hmin) cscale = 255.0 / (hmax - hmin);
  else cscale = 0;
  for (y=y1;y<y2;y++) {
   yy = (int)((y-y1)*ysf);
   for (x=x1;x<x2;x++) {
    xx = (int)((x-x1)*xsf);
    if (cscale == 0) c = 128;
    else c = (int) ((El(hf.a,xx,yy) - hmin) * cscale);
    if (c == TCOLOR) c--;
    GrPlot(x,y,c);
   } /* for y */
  } /* for x */
 } /* else (not hf.c) */
 return 0;
} /* h_plots() */

int ss_flat(int show4, int xgsize, int ygsize)
{
 int xoff;
 hfield hf;
 int x1,y1,x2,y2;
 char label[80];   /* HF text label */
 int tht = 13; /* height of screen label text */

 xoff = (xgsize-ygsize)/2;
 /* printf("res %dx%d\n",xgsize,ygsize); */
 
 if (show4) { /* show four matrices */
   DECSTACK;
   DECSTACK;
   DECSTACK;
   x1 = ygsize/2 + xoff; y1 = ygsize/2;
   x2 = ygsize + xoff; y2 = ygsize;
   hf = hst[hp];
   if (hf.a != NULL) {
    h_plots(hf,x1,y1,x2,y2);   /* fourth matrix */
    if (!(takefile && !echo_cmd)) {
     GrTextXY(x2+2,y1,hf.name,TCOLOR,0);
     sprintf(label,"%dx%d",hf.xsize,hf.ysize);
     GrTextXY(x2+2,y1+tht,label,TCOLOR,0);
     sprintf(label,"%1.2f|%1.2f",hf.min,hf.max);
     GrTextXY(x2+2,y1+tht*2,label,TCOLOR,0);
    }
   }
   INCSTACK;
   hf = hst[hp];
   if (hf.a != NULL) {
    h_plots(hf,xoff,y1,x1,y2);  /* third matrix */
    if (!(takefile && !echo_cmd)) {
     GrTextXY(2,y1,hf.name,TCOLOR,0);
     sprintf(label,"%dx%d",hf.xsize,hf.ysize);
     GrTextXY(2,y1+tht,label,TCOLOR,0);
     sprintf(label,"%1.2f|%1.2f",hf.min,hf.max);
     GrTextXY(2,y1+tht*2,label,TCOLOR,0);
    }
   }
   INCSTACK;
   hf = hst[hp];
   if (hf.a != NULL) {
    h_plots(hf,x1,0,x2,y1);    /* second matrix */
    if (!(takefile && !echo_cmd)) {
     GrTextXY(x2+2,5,hf.name,TCOLOR,0);
     sprintf(label,"%dx%d",hf.xsize,hf.ysize);
     GrTextXY(x2+2,5+tht,label,TCOLOR,0);
     sprintf(label,"%1.2f|%1.2f",hf.min,hf.max);
     GrTextXY(x2+2,5+tht*2,label,TCOLOR,0);
    }
   }
   INCSTACK;
   hf = hst[hp];
   if (hf.a != NULL) {
    h_plots(hf,xoff,0,x1,y1);   /* first matrix */
    if (!(takefile && !echo_cmd)) {
     GrTextXY(2,5,hf.name,TCOLOR,0);
     sprintf(label,"%dx%d",hf.xsize,hf.ysize);
     GrTextXY(2,5+tht,label,TCOLOR,0);
     sprintf(label,"%1.2f|%1.2f",hf.min,hf.max);
     GrTextXY(2,5+tht*2,label,TCOLOR,0);
    }
   }
 } else {  /* show just one matrix */
   hf = hst[hp];
   h_plots(hf,0,0,xgsize,ygsize);
   if (!(takefile && !echo_cmd)) {
     GrTextXY(2,5,hf.name,TCOLOR,0);
     sprintf(label,"%dx%d",hf.xsize,hf.ysize);
     GrTextXY(2,5+tht,label,TCOLOR,0);
     sprintf(label,"%1.2f|%1.2f",hf.min,hf.max);
     GrTextXY(2,5+tht*2,label,TCOLOR,0);
   }
 }
 return(0);
} /* ss_flat() */
